# coding=utf-8
# Copyright © 2015 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

"""
Moab Web Services client library

This module provides an API to access the REST interface of the Moab Web
Services. The REST interface is documented here:
    http://docs.adaptivecomputing.com/suite/8-0/enterprise/help.htm#topics/moabWebServices/0-intro/productOverview.htm

Create an instance of MWSClient with the URL and credentials for Moab Web
Services.  MWSClient follows a naming convention that maps method names to
REST URLs. For example, to get a Job, call the jobGet() method.  Returned
values are typically Python dictionaries converted directly from the JSON
response from the web server.

Example:

>>> from acpython.mwsclient import MWSClient
>>> client = MWSClient()
>>> nodes = client.nodeGetAll()
>>> len(nodes)
42
>>> nodes[0]['name']
u'r01idpx01'
>>> node = client.nodeGet('r01idpx01')
>>> node['states']['state']
u'Idle'
"""

import ConfigParser
import fnmatch
import logging
import urllib
import socket
import time
import json
import warnings

import requests
import requests.exceptions

import acpython.config
import acpython.version


logger = logging.getLogger(__name__)

DEFAULT_MWS_URL = "http://localhost:8080/mws"


class MWSException(Exception):

    message = 'MWS Exception'

    def __init__(self, message=None, server_messages=None, http_status=-1,
                 json_obj=None):
        Exception.__init__(self, message)
        if server_messages is None:
            server_messages = []
        if message:
            self.message = message
        self.server_messages = server_messages
        self.http_status = http_status
        self.json_obj = json_obj

    def __str__(self):
        if self.message:
            result = [self.message]
        else:
            result = [self.__class__.__name__]
        if self.http_status and self.http_status != -1:
            result.extend(['\n    HTTP status: ', str(self.http_status)])
        if self.server_messages:
            for server_message in self.server_messages:
                result.extend(['\n    ', server_message])
        return ''.join(result)


class MWSSetupException(MWSException):
    message = "Error while connecting to server"


class MWSNotImplemented(MWSException):
    message = "Not yet implemented in python-mws client"


class MWSValueError(MWSException):
    message = "Invalid parameter passed to MWSClient"


class MWSBadRequest(MWSException):
    message = "Server Responded with 400 Bad Request"


class MWSUnauthorized(MWSException):
    message = "Not authorized (bad username or password?)"


class MWSForbidden(MWSException):
    message = "You are not authorized to access this area."


class MWSNotFound(MWSException):
    message = "Server Responded with 404 Not Found"


class MWSMethodNotAllowed(MWSException):
    message = "Server Responded with 405 Method Not Allowed"


class MWSServerError(MWSException):
    message = "Server Responded with 500 Server Error"


def _create_exception_with_new_message(e, message):
    e.message = message
    return e


class MWSClient(object):
    def __init__(self,
                 url=None,
                 username=None,
                 password=None,
                 access_token=None,
                 token_type=None,
                 verbose=None,
                 no_connect_check=None,
                 disable_ssl_certificate_validation=None,
                 timeout=None,
                 client_id=None,
                 client_secret=None,
                 enable_version_warning=None,
                 enable_load_config=True,
                 ):
        """
        Create an MWS client object.

        If enable_load_config is True (the default), then the
        acpython.config.getConfig() function is used to fetch configuration
        info. The parameters to this method override config variables of the
        same name in the [mws] section of the config file.

        If access_token and token_type are given, then the constructor
        will return without attempting authentication. Otherwise,
        if the username and password parameters are given, they will
        be used to attempt authentication. If all four are not given
        (username, password, access_token, token_type), then the values
        in the config file will be used.

        If at least the username and password parameters are given (or
        access_token and token_type are given), then it is not considered
        an error if the config file could not be loaded. Otherwise, if the
        config file does not exist or cannot be read, an EnvironmentError
        exception is raised, containing the default config file location
        and suggested contents.

        If both client_id and client_secret are given, then OAuth
        authentication is used, otherwise, MWSClient uses basic authentication.

        :param str url: The MWS base URL. Default: 'http://localhost:8080/mws'
                        This URL should not end with a '/' character.

        :param str username: The user name for MWS authentication.
                             Required to be given as a parameter or in the
                             config file. No default.

        :param str password: The password for MWS authentication.
                             Required to be given as a parameter or in the
                             config file. No default.
        :param str access_token: The access token for oauth
        :param str token_type: The token type for oauth
        :param int verbose: An integer that controls request logging.
                            0: Don't log requests or responses. (the default)
                            1: Create 1-line debug log messages for every
                               request and response.
                            2: In addition to the level 1 logging, log the
                               entire request and response contents at debug
                               log level.

        :param bool no_connect_check: False (the default) to connect to MWS
                                 immediately and fetch version info,
                                 True to not fetch nor check MWS version info.

        :param bool disable_ssl_certificate_validation:
                           If False (the default), verify the server's
                           certificate when using the https protocol.

        :param float timeout: Number of seconds to wait for the server to
                            send bytes before raising an exception.
                            Default (None) is to wait without timing out.

        :param str client_id: (optional) client ID for OAuth authentication.

        :param str client_secret: (optional) client secret for OAuth.

        :param bool enable_version_warning: True (the default) to log a warning
                            message if the MWS server version doesn't match
                            what the client supports.

        :param bool enable_load_config: True (the default) to attempt to load
                            the acpython config file using the
                            acpython.config.loadConfig() function.
        """

        self.config = ConfigParser.SafeConfigParser()
        if enable_load_config:
            try:
                self.config = acpython.config.getConfig()
            except EnvironmentError:
                if (username and password) or (access_token and token_type):
                    pass
                else:
                    raise

        self.url = self._get_config_str('mws', 'url', override=url, default=DEFAULT_MWS_URL)
        if not self.url:
            raise MWSValueError("Missing parameter: url")

        self.username = self._get_config_str('mws', 'username', override=username)
        self.password = self._get_config_str('mws', 'password', override=password)
        self.access_token = self._get_config_str('mws', 'access_token', override=access_token)
        self.token_type = self._get_config_str('mws', 'token_type', override=token_type)
        self.client_id = self._get_config_str('mws', 'client_id', override=client_id)
        self.client_secret = self._get_config_str('mws', 'client_secret', override=client_secret)
        self.token_expires_in = None

        if (self.username and self.password) or (self.access_token and self.token_type):
            pass
        else:
            raise MWSValueError("Must provide username and password (or access_token and token_type)")

        self.verbose = self._get_config_int('mws', 'verbose', override=verbose)
        no_connect_check = self._get_config_bool('mws', 'no_connect_check', override=no_connect_check)
        disable_ssl_certificate_validation = self._get_config_bool(
            'mws', 'disable_ssl_certificate_validation',
            override=disable_ssl_certificate_validation)
        self.timeout = self._get_config_float('mws', 'timeout', override=timeout)
        if not self.timeout:
            self.timeout = None
        enable_version_warning = self._get_config_bool(
            'mws', 'enable_version_warning',
            override=enable_version_warning, default=True)

        self._auth_type = None
        self._session = requests.Session()
        self._session.verify = not disable_ssl_certificate_validation

        if self.token_type and self.access_token:
            self._auth_type = 'oauth'
            return

        if self.client_id and self.client_secret:
            self._do_oauth_authentication()
            self._auth_type = 'oauth'
        else:
            self._session.auth = (self.username, self.password)
            self._auth_type = 'basic'

        # Check the MWS version for compatibility with this client,
        # unless the connection check has been specifically disabled.
        if no_connect_check:
            self.about = {}
            self.version_ok = True
            return
        # If getAbout() succeeds without an exception, then we were able to
        # connect to a real MWS server.
        self.about = self.getAbout()
        if self.verbose:
            logger.info("MWS version: %s", self.about.get('version'))
            logger.info("MWS build number: %s", self.about.get('build'))
        # Check MWS version against acpython.version.moab_versions
        self.version_ok = False
        version = self.about.get('version')
        if version is not None:
            for pattern in acpython.version.moab_versions:
                if fnmatch.fnmatch(version, pattern):
                    self.version_ok = True
                    break
        if not self.version_ok:
            if enable_version_warning:
                logger.warn(
                    'MWS version "%s" does not match supported versions %r',
                    version, acpython.version.moab_versions)

    def _get_config_str(self, section, option, override=None, default=''):
        return self._get_config_obj(self.config.get, section, option,
                                    override, default)

    def _get_config_int(self, section, option, override=None, default=0):
        return self._get_config_obj(self.config.getint, section, option,
                                    override, default)

    def _get_config_float(self, section, option, override=None, default=0.0):
        return self._get_config_obj(self.config.getfloat, section, option,
                                    override, default)

    def _get_config_bool(self, section, option, override=None, default=False):
        return self._get_config_obj(self.config.getboolean, section, option,
                                    override, default)

    @staticmethod
    def _get_config_obj(function, section, option, override, default):
        if override is not None:
            return override
        try:
            return function(section, option)
        except ConfigParser.Error:
            return default

    def _do_oauth_authentication(self):
        params = {
            'grant_type': 'password',
            'username': self.username,
            'password': self.password,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }
        res, obj = self._make_request('/oauth/token', method='POST',
                                      body=params, body_is_json=False)
        self.access_token = obj['access_token']
        self.token_type = obj['token_type']
        self.token_expires_in = obj['expires_in']

    def _make_request(self, path, method='GET', body=None, headers=None,
                      body_is_json=True, url_params=None):
        """
        Execute an HTTP request. This is an *internal* method.
        :param path: Resource URI, relative to the MWS url (self.url).
                     self.url must *not* end with '/',
                     path *must* begin with '/'.
        :param method: HTTP method, one of 'GET', 'POST', 'PUT', or 'DELETE'.
        :param body: The contents of the request to send to MWS.
                     If body_is_json is True (the default), then this should
                     be an object (Python list or dictionary) that is
                     serializable to JSON. If body_is_json is False, then
                     this should be a byte string. For a GET request, don't
                     use this parameter at all.
        :param headers: (optional) A dictionary containing extra HTTP headers
                        to send to MWS.
        :param body_is_json: If True (the default), then the body parameter,
                             if given, is expected to be a list or dictionary,
                             and is serialized to JSON.
        :param url_params: (optional) A dictionary of query parameters to add
                           to the request URL. Note: you do not have to add
                           {'api-version': 3}, that is done for you already.
        :return: res, json_obj

            res is a response object with a status attribute, containing an
            integer with the HTTP status. res.status will be a number in the
            200's, since any other type of response will result in an
            exception being raised.

            json_obj is the returned data decoded from JSON format, or None.

        If anything goes wrong with the HTTP or network communication, or MWS
        returns an error, or a parameter check fails, then this method raises
        an exception derived from MWSException. MWSNotFound is raised in the
        case of a 404 error (resource not found).
        """
        allowed_methods = ('GET', 'POST', 'PUT', 'DELETE')
        if method not in allowed_methods:
            raise MWSValueError("Unsupported HTTP method: {0}".format(method))
        if headers is None:
            headers = {}
        headers['user-agent'] = 'python-mws'
        if self._auth_type == 'oauth':
            headers['Authorization'] = self.token_type + ' ' + self.access_token
        if body_is_json and body is not None:
            try:
                body = json.dumps(body)
                headers['Content-Type'] = 'application/json'
            except TypeError:
                raise MWSValueError("Data provided is not JSON-serializable")
        elif not body_is_json:
            if isinstance(body, unicode):
                body = body.encode('utf-8')
        if url_params is None:
            my_url_params = {'api-version': 3}
        else:
            my_url_params = url_params.copy()
            my_url_params['api-version'] = 3
        url = "{0}{1}".format(self.url, path)
        if self.verbose:
            request_size = 0
            if body is not None:
                request_size = len(body)
            logger.debug("Request: %s %s size=%d", method, url, request_size)
        if self.verbose > 1 and body is not None:
            logger.debug("Request body: %r", body)
        try:
            res = self._session.request(method, url, headers=headers,
                                        params=my_url_params,
                                        data=body, timeout=self.timeout)
        except socket.error as e:
            raise MWSSetupException(
                'Could not connect to url "%s" as user "%s": %s' % (
                    url, self.username, e))
        except requests.exceptions.Timeout as e:
            raise MWSSetupException(
                'Timeout while connecting to url "%s" as user "%s": %s' % (
                    url, self.username, e))
        except requests.exceptions.RequestException as e:
            raise MWSSetupException(
                'Error while connecting to url "%s" as user "%s": %s' % (
                    url, self.username, e))
        if self.verbose:
            logger.debug("Response: status=%s", res.status_code)
        # MWS can return useful JSON in the response body, even if the
        # status code is 40X or 50X. So we try to decode the content first,
        # before checking the status code.
        content_type = res.headers.get('content-type')
        json_obj = None
        messages = None
        if content_type is not None:
            content_type = content_type.split(';')[0]
            if content_type == 'application/json':
                json_obj = res.json()
                if isinstance(json_obj, dict):
                    messages = json_obj.get('messages')
                    if not messages:
                        message = json_obj.get('message')
                        if message:
                            messages = [message]
        if self.verbose > 1:
            logger.debug("Response json: %r", json_obj)
        if res.status_code == 400:
            raise MWSBadRequest(server_messages=messages, http_status=400, json_obj=json_obj)
        if res.status_code == 401:
            raise MWSUnauthorized(server_messages=messages, http_status=401)
        if res.status_code == 403:
            raise MWSForbidden(server_messages=messages, http_status=403)
        if res.status_code == 404:
            raise MWSNotFound(server_messages=messages, http_status=404, json_obj=json_obj)
        if res.status_code == 405:
            raise MWSMethodNotAllowed(server_messages=messages, http_status=405)
        if res.status_code < 200 or res.status_code > 299:
            # This includes all of the 500 server error messages, and
            # also any strange or unsupported response code. There should not
            # be any status codes in the 300's because the request module
            # automatically handles redirects.
            raise MWSServerError(server_messages=messages,
                                 http_status=res.status_code,
                                 json_obj=json_obj)
        if content_type is None:
            raise MWSException('No Content-Type in response, status=%d' %
                               (res.status_code,))
        elif content_type != 'application/json':
            raise MWSException(
                "Unsupported Content-Type. Got %s" % content_type)
        # For backwards compatibility, put status attribute on response object.
        # Outside of this method, the caller should not care whether res is an
        # httplib2 response object or a requests response object. The caller
        # should only look at the status attribute, nothing else.
        res.status = res.status_code
        return res, json_obj

    def getAbout(self):
        aboutPaths = ('/rest/diag/about', '/rest/debug/about')
        for aboutPath in aboutPaths:
            try:
                res, data = self._make_request(aboutPath)
            except MWSNotFound:
                continue
            if res.status != 200:
                raise MWSException("Bad Return Code from %s: %s" % (
                    aboutPath, res.status))
            return data
        raise MWSSetupException(
            'No response from "about" paths on server: %s' % (aboutPaths,))

    def userIsAdmin(self):
        # We can't use _make_request because it expects that the content-type
        # is always application/json.
        headers = {}
        if self._auth_type == 'oauth':
            headers['Authorization'] = self.token_type + ' ' + self.access_token
        url = self.url + '/admin/diag'
        result = self._session.request(method='GET', url=url, headers=headers, timeout=self.timeout)
        if result.status_code == requests.codes.ok:
            return True
        if result.status_code == requests.codes.forbidden:
            return False
        raise MWSException("Unexpected return code: " + result.status_code)

    def logout(self):
        # We can't use _make_request because it expects that the content-type
        # is always application/json.
        headers = {}
        if self._auth_type == 'oauth':
            headers['Authorization'] = self.token_type + ' ' + self.access_token
        url = self.url + '/logout/index'
        self._session.request(method='GET', url=url, headers=headers, timeout=self.timeout)

    def credentialGet(self, credential_type, credential_name=None):
        """ Get credential information.
        :param str credential_type:
            One of users, groups, accounts, classes, qoses, or belongs-to
        :param str credential_name:
            The name of the credential to get. If None, get all credentials of the given type.
            Ignored if credential_type is belongs-to.
        :return: dict
        """

        path = '/rest/credentials/' + credential_type
        if credential_name and credential_type != 'belongs-to':
            path = path + '/' + credential_name
        res, data = self._make_request(path)
        return data

    def distinctGet(self, resource, field, query=None):
        """ Get distinct (unique) values from another MWS resource.
        :param str resource:
        :param str field:
        :param dict query:
        :return: dict
        """

        path = '/rest/distinct/%s/%s' % (resource, field)
        params = {}
        if isinstance(query, dict) and len(query) != 0:
            params['query'] = json.dumps(query)
        res, data = self._make_request(path=path, method='GET', url_params=params)
        return data

    ### Accessing Events ###

    def eventCreate(self, attributes=None, autoReload=True):
        if attributes is None:
            attributes = {}
        res, data = self._make_request('/rest/events', 'POST', attributes)
        if res.status == 201:
            if autoReload and not isinstance(data, list):
                return self.eventGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def eventGet(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('Event ID not specified')
        if id is None:
            id = attributes['id']
        try:
            res, data = self._make_request('/rest/events/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find event '%s'" % (id,))

    ### Accessing Images ###

    def imagesGetAll(self):
        res, data = self._make_request('/rest/images')
        return data['results']

    def imageGet(self, id):
        try:
            res, data = self._make_request('/rest/images/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find image '%s'" % (id,))

    def imageFind(self, query):
        if isinstance(query, dict):
            params = {'query': json.dumps(query)}
        else:
            params = {'query': query}
        res, data = self._make_request('/rest/images', url_params=params)
        return data

    # Alias for compatibility.
    imageGetByQuery = imageFind

    def imageCreate(self, attributes, autoReload=True):
        if not attributes:
            raise ValueError('attributes cannot be empty')
        res, data = self._make_request('/rest/images', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                return self.imageGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def imageModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('Image Id not specified')
        if id is None:
            id = attributes['id']
        try:
            res, data = self._make_request('/rest/images/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, 'Could not find node %s to modify' % id)

    def imageDelete(self, id):
        try:
            self._make_request('/rest/images/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find image '%s' to delete" % (id,))

    ### Accessing Jobs ###
    def jobGetAll(self):
        res, data = self._make_request('/rest/jobs')
        return data['results']

    def jobGetActive(self):
        res, data = self._make_request(
            '/rest/jobs', url_params={'query': '{"isActive": true}'})
        return data['results']

    def jobGetCompleted(self):
        res, data = self._make_request(
            '/rest/jobs', url_params={'query': '{"isActive": false}'})
        return data['results']

    def jobGet(self, id):
        try:
            res, data = self._make_request('/rest/jobs/%s' % (id,))
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find job '%s'" % (id,))

    def jobGetWithQuery(self, query, sort=None, maximum=None, offset=None, fields=None, exclude_fields=None):
        """ Get the list of jobs that match the given query.
        :param dict query:
        :param dict sort:
        :param int maximum:
        :param int offset:
        :param str fields:
        :param str exclude_fields:
        :return: dict
            The fields in the dict:
                resultCount int: The number of items returned, as limited by maximum and offset parameters
                totalCount int: The number of items in the entire result set
                results list of dict: The jobs that match the query
        For documentation on the meanings of the parameters, see these links:
        http://docs.adaptivecomputing.com/8-1-0/enterprise/help.htm#topics/moabWebServices/3-apiDocs/globalUrlParameters.htm
        http://docs.adaptivecomputing.com/8-1-0/enterprise/help.htm#topics/moabWebServices/4-resources/jobs.htm
        Some parameters have been renamed slightly to conform to Python naming.
            max -> maximum
            exclude-fields -> exclude_fields
        """
        params = {}
        if isinstance(query, dict) and len(query) != 0:
            params['query'] = json.dumps(query)
        if isinstance(sort, dict) and len(sort) != 0:
            params['sort'] = json.dumps(sort)
        if isinstance(maximum, int):
            params['max'] = maximum
        if isinstance(offset, int):
            params['offset'] = offset
        if isinstance(fields, str):
            params['fields'] = fields
        if isinstance(exclude_fields, str):
            params['exclude-fields'] = exclude_fields
        res, data = self._make_request(path='/rest/jobs', method='GET', url_params=params)
        return data

    def jobModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'name' not in attributes:
            raise MWSValueError('Job name not specified')
        if id is None:
            id = attributes['name']
        try:
            res, data = self._make_request('/rest/jobs/%s' % (id,), 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find job '%s' to modify" % (id,))

    def jobCancel(self, id):
        try:
            self._make_request('/rest/jobs/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find job '%s' to delete" % (id,))

    def jobHold(self, name, body):
        """ Add hold to a job.
        :param str name: The job
        :param dict body: The body of the request.
        :return:
        """
        if not isinstance(body, dict):
            raise MWSValueError("Invalid body parameter: must be a dictionary.")
        if len(body) == 0:
            raise MWSValueError("Invalid body parameter: must not be empty.")
        res, data = self._make_request(
            path='/rest/jobs/%s/hold' % name,
            method='PUT',
            body=body)
        return data

    def jobRelease(self, name, body):
        """ Release hold on a job.
        :param str name: The job
        :param dict body: The body of the request. If None, all holds will be released.
        :return:
        """
        if body is None:
            body = {}
        if not isinstance(body, dict):
            raise MWSValueError("Invalid body parameter: must be a dictionary.")
        res, data = self._make_request(
            path='/rest/jobs/%s/unhold' % name,
            method='PUT',
            body=body)
        return data

    def jobUnhold(self, id):
        warnings.warn("Use jobRelease instead.", DeprecationWarning)
        try:
            self._make_request('/rest/jobs/%s/unhold' % id, 'PUT')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, 'Could not find job %s to unhold' % id)

    def jobCreate(self, attributes=None, autoReload=True, url_params=None):
        if attributes is None:
            attributes = {}
        res, data = self._make_request(
            path='/rest/jobs',
            method='POST',
            body=attributes,
            url_params=url_params)
        if res.status == 201:
            if autoReload:
                timeout_sec = 30.0
                check_interval = 2.0
                t0 = time.time()
                while (time.time() - t0) <= timeout_sec:
                    try:
                        return self.jobGet(data['name'])
                    except MWSNotFound:
                        time.sleep(check_interval)
                else:
                    raise MWSNotFound("Couldn't find job " + data['name'])
            else:
                return data
        if res.status == 202:
            return data

    def jobArrayCreate(self, body, url_params=None):
        if not isinstance(body, dict):
            raise MWSValueError("Invalid body parameter: must be a dictionary.")
        if len(body) == 0:
            raise MWSValueError("Invalid body parameter: must not be empty.")
        response, data = self._make_request(
            path='/rest/job-arrays',
            method='POST',
            body=body,
            url_params=url_params)
        return data

    ### Accessing Nodes ###

    def nodeGetAll(self):
        res, data = self._make_request('/rest/nodes')
        return data['results']

    def nodeGet(self, id):
        try:
            res, data = self._make_request('/rest/nodes/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find node %s" % id)

    def nodeModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'name' not in attributes:
            raise MWSValueError('Node name not specified')
        if id is None:
            id = attributes['name']
        try:
            res, data = self._make_request('/rest/nodes/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, 'Could not find node %s to modify' % id)

    def permissionCreate(self, attributes, autoReload=True):
        if not attributes:
            raise ValueError('attributes cannot be empty')
        res, data = self._make_request('/rest/permissions', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                return self.permissionGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def permissionGet(self, id):
        try:
            res, data = self._make_request('/rest/permissions/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find permission '%s'" % (id,))

    def permissionDelete(self, id):
        try:
            self._make_request('/rest/permissions/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find permission '%s' to delete" % (id,))

    def principalCreate(self, attributes, autoReload=True):
        if not attributes:
            raise ValueError('attributes cannot be empty')
        res, data = self._make_request('/rest/principals', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                return self.principalGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def principalGet(self, id):
        try:
            res, data = self._make_request('/rest/principals/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find principal '%s'" % (id,))

    def principalModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('principal Id not specified')
        if id is None:
            id = attributes['id']
        try:
            res, data = self._make_request('/rest/principals/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, 'Could not find principal %s to modify' % id)

    def principalDelete(self, id):
        try:
            self._make_request('/rest/principals/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find principal '%s' to delete" % (id,))

    ### Accessing Reservations ###
    def reservationGetAll(self):
        res, data = self._make_request('/rest/reservations')
        return data['results']

    def reservationGet(self, id):
        try:
            res, data = self._make_request('/rest/reservations/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find reservation '%s'" % (id,))

    def reservationModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('Reservation ID not specified')
        if id is None:
            id = attributes['id']
        try:
            res, data = self._make_request('/rest/reservations/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find reservation '%s' to modify" % (id,))

    def reservationRelease(self, id):
        try:
            self._make_request('/rest/reservations/%s' % id, 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find reservation '%s' to delete" % (id,))

    def reservationCreate(self, attributes=None, autoReload=True):
        if attributes is None:
            attributes = {}
        res, data = self._make_request('/rest/reservations', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                return self.reservationGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def roleCreate(self, attributes, autoReload=True):
        if not attributes:
            raise ValueError('attributes cannot be empty')
        res, data = self._make_request('/rest/roles', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                return self.roleGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def roleGet(self, id):
        try:
            res, data = self._make_request('/rest/roles/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find role '%s'" % (id,))

    def roleModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('role Id not specified')
        if id is None:
            id = attributes['id']
        try:
            res, data = self._make_request('/rest/roles/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, 'Could not find role %s to modify' % id)

    def roleDelete(self, id):
        try:
            self._make_request('/rest/roles/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find role '%s' to delete" % (id,))

    ### Accessing VMs ###

    def vmGetAll(self):
        res, data = self._make_request('/rest/vms')
        return data['results']

    def vmGet(self, vmId):
        res, data = self._make_request('/rest/vms/%s' % (vmId,))
        return data

    def vmModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'name' not in attributes:
            raise MWSValueError('VM name not specified')
        if id is None:
            id = attributes['name']
        try:
            res, data = self._make_request('/rest/vms/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find VM '%s' to modify" % id)

    def vmMigrate(self, vmId, nodeId):
        """
        Migrate VM vmId to node nodeId
        """
        attributes = {'host': {'name': nodeId}}
        return self.vmModify(vmId, attributes)

    def vmDestroy(self, id):
        try:
            self._make_request('/rest/vms/%s' % id, 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Could not find VM '%s' to delete" % id)

    ### Accessing Services ###

    def serviceGetAll(self):
        res, data = self._make_request('/rest/services')
        return data['results']

    def serviceGet(self, name_or_id, show_recursive_vc=True):
        # There can be spaces in a service name
        quoted_name_or_id = urllib.quote(name_or_id)
        url = '/rest/services/%s' % (quoted_name_or_id,)
        params = {}
        if show_recursive_vc:
            params['show-recursive-vc'] = 'true'
        try:
            res, data = self._make_request(url, url_params=params)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find service '%s'" % (name_or_id,))

    def serviceFind(self, query, show_recursive_vc=True):
        if isinstance(query, dict):
            params = {'query': json.dumps(query)}
        else:
            params = {'query': query}
        if show_recursive_vc:
            params['show-recursive-vc'] = 'true'
        res, data = self._make_request('/rest/services', url_params=params)
        return data

    def serviceCreate(self, attributes=None, autoReload=True):
        if attributes is None:
            attributes = {}
        res, data = self._make_request('/rest/services', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                if 'id' in data:
                    return self.serviceGet(data['id'])
                return self.serviceGet(data['name'])
            else:
                return data
        if res.status == 202:
            return data

    def serviceUpdateField(self, id, *params, **kwparams):
        """
        For backwards compatibility, this method may be called two different
        ways:

        client.serviceUpdateField(id, fieldName, attributes, autoReload=True)
        client.serviceUpdateField(id, body, autoReload=True)

        If you use the second form, set body = { fieldName: attributes }

        If you pass autoReload, please use a keyword parameter, not a
        positional parameter.
        """
        autoReload = kwparams.get('autoReload', True)
        if len(params) == 1:
            body = params[0]
        elif len(params) == 2:
            if type(params[0]) == dict and type(params[1]) == bool:
                # Caller disregarded my advice and passed autoReload
                # positionally.
                body, autoReload = params
            else:
                fieldName, attributes = params
                body = {fieldName: attributes}
        elif len(params) == 3:
            fieldName, attributes, autoReload = params
            body = {fieldName: attributes}
        else:
            raise ValueError('Invalid number of parameters')
        res, data = self._make_request(
            '/rest/services/' + urllib.quote(id), 'PUT', body)
        if autoReload:
            return self.serviceGet(data['id'])
        return data

    def serviceUpdateSharedDataField(self, id, fieldName, attributes,
                                     autoReload=True):
        body = {"attributes": {"sharedData": {fieldName: attributes}}}
        res, data = self._make_request(
            '/rest/services/' + urllib.quote(id), 'PUT', body)
        if autoReload:
            return self.serviceGet(data['id'])
        return data

    def serviceModify(self, name_or_id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if name_or_id is None:
            name_or_id = attributes.get('id')
        if name_or_id is None:
            name_or_id = attributes.get('name')
        if name_or_id is None:
            raise MWSValueError('Service name or ID not specified')
        # There can be spaces in a service name
        quoted_name_or_id = urllib.quote(name_or_id)
        url = '/rest/services/%s' % (quoted_name_or_id,)
        try:
            res, data = self._make_request(url, 'PUT', attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Service '%s' to modify" % (id,))

    def serviceDelete(self, name_or_id, force=False):
        # There can be spaces in a service name
        quoted_name_or_id = urllib.quote(name_or_id)
        url = '/rest/services/%s' % (quoted_name_or_id,)
        url_params = None
        if force:
            url_params = {'reallydelete': 'True'}
        try:
            self._make_request(url, 'DELETE', url_params=url_params)
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find service '%s' to delete" % (name_or_id,))

    ### Accessing Service Templates ###

    def serviceTemplateGetAll(self):
        res, data = self._make_request('/rest/service-templates')
        return data['results']

    def serviceTemplateGet(self, name_or_id):
        # There can be spaces in a service template name
        quoted_name_or_id = urllib.quote(name_or_id)
        url = '/rest/service-templates/%s' % (quoted_name_or_id,)
        try:
            res, data = self._make_request(url)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find service template '%s'" % (name_or_id,))

    def serviceTemplateCreate(self, attributes=None, autoReload=True):
        if attributes is None:
            attributes = {}
        res, data = self._make_request('/rest/service-templates', 'POST',
                                       attributes)
        if res.status == 201:
            if autoReload:
                return self.serviceTemplateGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def serviceTemplateModify(self, name_or_id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if name_or_id is None:
            name_or_id = attributes.get('id')
        if name_or_id is None:
            name_or_id = attributes.get('name')
        if name_or_id is None:
            raise MWSValueError('Service Template name or ID not specified')
        # There can be spaces in a service name
        quoted_name_or_id = urllib.quote(name_or_id)
        url = '/rest/service-templates/%s' % (quoted_name_or_id,)
        try:
            res, data = self._make_request(url, 'PUT', attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Service Template '%s' to modify" % (id,))

    def serviceTemplateDelete(self, name_or_id):
        # There can be spaces in a service template name
        quoted_name_or_id = urllib.quote(name_or_id)
        url = '/rest/service-templates/%s' % (quoted_name_or_id,)
        try:
            self._make_request(url, 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find service template '%s' to delete" % (
                    name_or_id,))

    ### Accessing Plugin Web Services ###

    def pluginService(self, plugin, plugin_service, method='GET',
                      url_params=None, body=None):
        url = '/rest/plugins/%s/services/%s' % (plugin, plugin_service)
        result, data = self._make_request(url, method, url_params=url_params,
                                          body=body)
        return data

    ### Accessing Plugin Types ###

    def pluginTypeGetAll(self):
        res, data = self._make_request('/rest/plugin-types')
        return data['results']

    def pluginTypeGet(self, id):
        url = '/rest/plugin-types/%s' % (id,)
        try:
            res, data = self._make_request(url)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Plugin Type '%s'" % (
                    id,))

    def pluginTypePUT(self, plugin_type_bytes, plugin_is_jar=False):
        if plugin_is_jar:
            content_type = 'application/x-jar'
        else:
            content_type = 'text/plain'
        res, data = self._make_request('/rest/plugin-types', 'PUT',
                                       plugin_type_bytes,
                                       headers={'Content-Type': content_type},
                                       body_is_json=False)
        if res.status in [200, 201, 202]:
            return data

    def pluginTypeCreate(self, plugin_type_bytes, plugin_is_jar=False):
        """
        Create a new Plugin Type
        plugin_type_bytes:
            If plugin_is_jar, the bytes of the plugin Jar file, otherwise,
            the characters (can be unicode) of the Groovy source file.
        """
        # Create and update both use PUT
        return self.pluginTypePUT(plugin_type_bytes,
                                  plugin_is_jar=plugin_is_jar)

    def pluginTypeUpdate(self, plugin_type_bytes, plugin_is_jar=False):
        """
        Update an existing Plugin Type
        plugin_type_bytes:
            If plugin_is_jar, the bytes of the plugin Jar file, otherwise,
            the characters (can be unicode) of the Groovy source file.
        """
        # Create and update both use PUT
        return self.pluginTypePUT(plugin_type_bytes,
                                  plugin_is_jar=plugin_is_jar)

    ### Accessing Plugin instances ###

    def pluginCreate(self, id, pluginType, pollInterval=30, autoStart=True,
                     config=None):
        """
        Create a new Plugin instance.
        config: A dictionary of config parameters for the plugin, or None
        """
        attributes = {
            'id': id,
            'pluginType': pluginType,
            'pollInterval': pollInterval,
            'autoStart': autoStart,
        }
        if config is not None:
            attributes['config'] = config
        res, data = self._make_request('/rest/plugins', 'POST', attributes)
        return data

    def pluginGetAll(self):
        res, data = self._make_request('/rest/plugins')
        return data['results']

    def pluginGet(self, id):
        url = '/rest/plugins/%s' % (id,)
        try:
            res, data = self._make_request(url)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Plugin '%s'" % (id,))

    def pluginUpdate(self, id, pollInterval=30, autoStart=True, state='STARTED',
                     config=None, **kwparams):
        """
        Update an existing Plugin instance.
        config: A dictionary of config parameters for the plugin, or None

        >>> # Update just one attribute of a plugin, leaving the rest alone
        >>> client = MWSClient()
        >>> plugin = client.pluginGet('myPlugin')
        >>> plugin['pollInterval'] = 60
        >>> client.pluginUpdate('myPlugin', **plugin)
        """
        attributes = {
            'pollInterval': pollInterval,
            'autoStart': autoStart,
            'state': state,
        }
        if config is not None:
            attributes['config'] = config
        attributes.update(kwparams)
        for attrname in ('id', 'pluginType'):
            if attrname in attributes:
                del attributes[attrname]
        try:
            res, data = self._make_request('/rest/plugins/%s' % (id,), 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Plugin '%s' to modify" % id)

    def pluginDelete(self, id):
        try:
            self._make_request('/rest/plugins/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Plugin '%s' to delete" % (id,))

    def tenantCreate(self, attributes, autoReload=True):
        if not attributes:
            raise ValueError('attributes cannot be empty')
        res, data = self._make_request('/rest/tenants', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                return self.tenantGet(data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def tenantGet(self, id):
        try:
            res, data = self._make_request('/rest/tenants/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find tenant '%s'" % (id,))

    def tenantModify(self, id=None, attributes=None):
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('tenant Id not specified')
        if id is None:
            id = attributes['id']
        try:
            res, data = self._make_request('/rest/tenants/%s' % id, 'PUT',
                                           attributes)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, 'Could not find tenant %s to modify' % id)

    def tenantDelete(self, id):
        try:
            self._make_request('/rest/tenants/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find tenant '%s' to delete" % (id,))

    ### Accessing Virtual Containers ###

    def vcGetAll(self):
        """Get all Virtual Containers"""
        res, data = self._make_request('/rest/vcs')
        return data['results']

    def vcGet(self, id):
        """Get a single Virtual Container"""
        try:
            res, data = self._make_request('/rest/vcs/%s' % id)
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Virtual Container '%s'" % (id,))

    def vcCreate(self, attributes=None, autoReload=True):
        """Create a Virtual Container"""
        if attributes is None:
            attributes = {}
        res, data = self._make_request('/rest/vcs', 'POST', attributes)
        if res.status == 201:
            if autoReload:
                timeout_sec = 30.0
                check_interval = 2.0
                t0 = time.time()
                while (time.time() - t0) <= timeout_sec:
                    try:
                        return self.vcGet(data['id'])
                    except MWSNotFound:
                        time.sleep(check_interval)
                else:
                    raise MWSNotFound(
                        "Couldn't find Virtual Container '%s'" % data['id'])
            else:
                return data
        if res.status == 202:
            return data

    def vcModify(self, id=None, attributes=None, change_mode=None):
        """Modify a Virtual Container
        change_mode parameter must be 'add', 'remove', or 'set'
        """
        if attributes is None:
            attributes = {}
        if id is None and 'id' not in attributes:
            raise MWSValueError('Virtual Container ID not specified')
        if id is None:
            id = attributes['id']
        valid_change_modes = ('add', 'remove', 'set')
        if not change_mode in valid_change_modes:
            raise MWSValueError(
                "change_mode %r is not one of the valid values: %r" % (
                    change_mode, valid_change_modes))
        try:
            res, data = self._make_request(
                '/rest/vcs/%s' % (id,), 'PUT',
                attributes,
                url_params={'change-mode': change_mode})
            return data
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Virtual Container '%s' to modify" % (id,))

    def vcDestroy(self, id):
        try:
            self._make_request('/rest/vcs/%s' % (id,), 'DELETE')
            return True
        except MWSNotFound, e:
            raise _create_exception_with_new_message(
                e, "Couldn't find Virtual Container '%s' to delete" % (id,))

    ### Insight support ###

    def insightGet(self, view_name,
                   scope='priv',
                   fields=None,
                   group_by=None,
                   having=None,
                   max=None,
                   offset=None,
                   pretty=None,
                   query=None,
                   sort=None):
        """Perform a query against an Insight view.

        Notes about the having, query, and sort parameters:
            Each of these parameters should either be a JSON object in the
            form of a Python dictionary, or None.
            * Do not pass a string for any of these parameters.
            * Do not perform any URL quoting (with "%" characters, etc.) on
              any values within these parameters. This method automatically
              does URL quoting, conversion from unicode to UTF-8, etc.

        Notes about fields and group_by parameters:
            Each of these parameters should be a either a sequence (list
            or tuple or generator) of strings, or None. Do not pass a string
            for either of these parameters.

        :param str view_name: The name of an Insight table or view

        :param str scope: The scope of the query. Valid values are:
                'priv' (the default): Get all rows in a view
                'user': Get all rows the current user can see

        :param sequence fields: (optional) a sequence of field names.
                Includes only specified fields in output.

        :param sequence group_by: (optional) a sequence of group-by item names.
                Used in conjunction with aggregate functions.

        :param dict having: (optional) A dictionary containing a "having"
                clause. This is in the same format as the query parameter,
                but it queries for specific rows using results from aggregate
                functions.

        :param int max: (optional) The maximum number of items to return.

        :param int offset: (optional) The index of the first item to return.

        :param bool pretty: (optional, default False) True to instruct server
                to return nicely formatted JSON. Caller of this method won't
                notice any difference because this method returns an object,
                not a string. Only useful when tracing bytes over network.

        :param dict query: (optional) A dictionary containing a database query.
                It is possible to query by one or more fields based on a subset
                of MongoDB query syntax. (See the Insight database docs.)

        :param dict sort: (optional) A dictionary with keys that are column
                names to sort by, and values indicating sort direction.
                Use 1 for ascending and -1 for descending.
                Example: sort={"column-name": -1}

        :return dict: The JSON results from the server. Example:

            {
                u'totalCount': 89,
                u'resultCount': 3,
                u'results': [
                    { ... object 1 ...},
                    { ... object 2 ...},
                    { ... object 3 ...},
                ]
            }
        """
        if not view_name or not isinstance(view_name, (str, unicode)):
            raise MWSValueError("Invalid view_name parameter")
        if (not isinstance(scope, (str, unicode)) or
                scope not in ('priv', 'user')):
            raise MWSValueError("Invalid scope parameter: must be priv or user")
        path = "/rest/insight/{0}/{1}".format(scope, view_name)
        params = {}

        def _add_sequence_param(name, param):
            if param is None:
                return
            if isinstance(param, (str, unicode)):
                raise MWSValueError(
                    "Invalid {0} parameter: must not be a string".format(name))
            list_obj = []
            for item in param:
                if isinstance(item, unicode):
                    item = item.encode('utf-8')
                if not isinstance(item, str):
                    raise MWSValueError(
                        "Item in {0} is not a string".format(name))
                if ',' in item:
                    raise MWSValueError(
                        "Comma not allowed in item in {0}".format(name))
                list_obj.append(item)
            if not list_obj:
                # Don't send empty list
                return
            params[name] = ','.join(list_obj)

        def _add_dict_param(name, param):
            if param is None:
                return
            if not isinstance(param, dict):
                raise MWSValueError(
                    "Invalid {0} parameter: must be a dict".format(name))
            if not param:
                # Don't send empty dictionary
                return
            param_str = json.dumps(param)
            params[name] = param_str

        def _add_int_param(name, param):
            if param is None:
                return
            if not isinstance(param, int):
                raise MWSValueError(
                    "Invalid {0} parameter: must be an int".format(name))
            params[name] = param

        def _add_bool_param(name, param):
            if param is None:
                return
            if not isinstance(param, bool):
                raise MWSValueError(
                    "Invalid {0} parameter: must be a bool".format(name))
            params[name] = param

        _add_sequence_param('fields', fields)
        _add_sequence_param('group_by', group_by)
        _add_dict_param('having', having)
        _add_int_param('max', max)
        _add_int_param('offset', offset)
        _add_bool_param('pretty', pretty)
        _add_dict_param('query', query)
        _add_dict_param('sort', sort)

        res, data = self._make_request(path, url_params=params)
        return data
