# coding=utf-8
# Copyright © 2015 by Adaptive Computing Enterprises, Inc. All Rights Reserved.

"""
acpython utility class and functions
"""

import errno
import fcntl
import os
import stat
import time


def cacheFileIsStale(filename, interval):
    """
    Return True if the cache file needs to be refreshed.

    Details:
    If the file does not exist, or is zero-length, or was modified further
    than interval seconds in the past, or if the modification time itself
    is in the future (e.g. because the clock has been changed), return
    True.  Otherwise, return False. Note: file modification times may only
    be accurate to 1-second resolution.
    """
    try:
        s = os.stat(filename)
    except OSError:
        # Cannot stat, assume does not exist
        return True
    if s.st_size == 0:
        return True
    age = time.time() - s.st_mtime
    if age > interval:
        return True
    if age < 0:
        return True
    return False


class LockedFile(object):

    """
    A file locked with advisory locking using fcntl.lockf
    
    Use cases:
     * One or more processes asynchronously writing to a file, another
       group of processes asynchronously reading from that file, all
       processes see consistent contents.
     * Multiple instances of a program all doing the same query and caching
       the results in the same file. Use cacheFileIsStale() to check the
       cache file age and re-run the query periodically.
    """

    def __init__(self, name, 
                 mode=stat.S_IRUSR|stat.S_IWUSR|stat.S_IRGRP|stat.S_IROTH,
                 bufsize=-1,
                 exclusive=False, 
                 timeout_sec=60, retry_interval_sec=0.1):
        """
        name:
            Name of file to lock. The file is created if it doesn't exist.

        mode:
            File permission mode bits, used if the file does not already
            exist.  Default is read and write for owner, read permission
            for group and others (octal 644).

        bufsize:
            0: unbuffered, 1: line buffered, any other positive value means
            use a buffer of (approximately) that size. If negative, use the
            system default buffer size.

        exclusive:
            True to get an exclusive lock, False for shared lock.  If an
            non-exclusive lock is held, writes are not allowed on this file
            object.

        timeout_sec:
            Number of seconds to keep trying to acquire lock.
            Raise an IOError on timeout waiting to acquire the lock.

        retry_interval_sec:
            Time interval in seconds to wait between attempts to acquire
            the loc.

        Call close() on this object to unlock the file.
        """
        self.name = name
        self.mode = mode
        self.bufsize = bufsize
        self.timeout_sec = timeout_sec
        self.retry_interval_sec = retry_interval_sec
        # Open the file, creating the file if it doesn't already exist.
        # The file object is always opened in read-write mode, even when
        # the lock is shared, so that the lock can be changed to exclusive
        # later.
        fd = os.open(name, os.O_CREAT|os.O_RDWR, mode)
        self.fileobj = os.fdopen(fd, 'r+', bufsize)
        self._lock(exclusive)

    def __del__(self):
        try:
            self.close()
        except:
            # close() may fail if __init__ didn't complete
            pass

    def changeLockType(self, exclusive):
        """
        Change the type of lock held on this file.

        exclusive: boolean - True to change the lock type to exclusive,
                             False to change it to shared.

        If the lock type is changed from non-exclusive to exclusive, then
        writes will be allowed on this object.
        Raises an IOError on timeout waiting to change the lock type.
        Typically, changing from exclusive to non-exclusive will succeed
        immediately.
        """
        self._lock(exclusive)

    def _lock(self, exclusive):
        fd = self.fileobj.fileno()
        if exclusive:
            op = fcntl.LOCK_EX
        else:
            op = fcntl.LOCK_SH
        endtime = time.time() + self.timeout_sec
        while True:
            try:
                # Lock exactly one byte at the start of the file.
                # This works even if the file is zero length, because
                # lockf() allows you to lock past the end of a file.
                fcntl.lockf(fd, op|fcntl.LOCK_NB, 1)
                self.exclusive = exclusive
                return
            except IOError as e:
                if e.errno not in (errno.EACCES, errno.EAGAIN):
                    raise e
            if time.time() >= endtime:
                raise IOError("Timed out waiting for lock on file %s" %
                              self.name)
            time.sleep(self.retry_interval_sec)

    def close(self):
        fileobj = self.fileobj
        if not fileobj:
            return
        self.fileobj = None
        fileobj.close()

    @property
    def closed(self):
        return (self.fileobj is None)

    def fileno(self):
        return self.fileobj.fileno()

    def flush(self):
        return self.fileobj.flush()

    def isatty(self):
        return self.fileobj.isatty()

    def next(self):
        return self.fileobj.next()

    def read(self, size=None):
        if size is None:
            return self.fileobj.read()
        return self.fileobj.read(size)

    def readable(self):
        return (self.fileobj is not None)

    def readline(self, size=None):
        if size is None:
            return self.fileobj.readline()
        return self.fileobj.readline(size)

    def readlines(self, sizehint=None):
        if sizehint is None:
            return self.fileobj.readlines()
        return self.fileobj.readlines(sizehint)

    def seek(self, offset, whence = 0):
        return self.fileobj.seek(offset, whence)

    def tell(self):
        return self.fileobj.tell()

    def truncate(self, size=None):
        if not self.exclusive:
            raise IOError("truncate() only allowed if exclusive lock held")
        if size is None:
            return self.fileobj.truncate()
        return self.fileobj.truncate(size)

    def write(self, data):
        if not self.exclusive:
            raise IOError("write() only allowed if exclusive lock held")
        return self.fileobj.write(data)

    def writelines(self, sequence):
        if not self.exclusive:
            raise IOError("write() only allowed if exclusive lock held")
        return self.fileobj.writelines(sequence)
